/*
** Module   : P2JField.java
** Abstract : Helper class for dynamic temporary tables.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 SVL 20130624 Created initial version.
*/

package com.goldencode.p2j.persist;

import antlr.*;
import com.goldencode.ast.*;
import com.goldencode.compile.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.xml.*;
import org.hibernate.cfg.*;
import org.hibernate.internal.*;
import org.hibernate.mapping.*;
import java.io.*;
import java.util.*;
import java.util.Map;
import java.util.concurrent.atomic.*;

/**
 * Helper class for dynamic temporary tables.
 */
public class DynamicTablesHelper
{
   /** Base package for generated DMOs. */
   public final static String DMO_BASE_PACKAGE = "com.goldencode.p2j.persist.dynamic";

   /** Category name for DMO implementation classes, used in conversion rules. */
   private final static String DMO_IMPL_CATEGORY = "impl";

   /** Category name for DMO interfaces, used in conversion rules. */
   private final static String DMO_IFACE_CATEGORY = "iface";

   /** Singleton instance. */
   private static volatile DynamicTablesHelper instance;

   /** Counter used for generation of unique names for dynamic tables. */
   private final AtomicLong tableCounter = new AtomicLong(0);

   /** Param type to Progress type. */
   private final static Map<ParmType, Integer> propTypeToProgress
                                                = new HashMap<ParmType, Integer>();

   /**
    * Mapping of actual field names specified in ADD-NEW-FIELD to Hibernate names ("field1",
    * "field2", ...).
    */
   private final static ContextLocal<Map<Class, Map<String, String>>> fields4GLtoHibernate=
         new ContextLocal<Map<Class, Map<String, String>>>()
   {
      protected Map<Class, Map<String, String>> initialValue()
      {
         return new HashMap<Class, Map<String, String>>();
      }
   };

   static
   {
      propTypeToProgress.put(ParmType.CHAR, ProgressParserTokenTypes.FIELD_CHAR);
      propTypeToProgress.put(ParmType.INT, ProgressParserTokenTypes.FIELD_INT);
      propTypeToProgress.put(ParmType.INT64, ProgressParserTokenTypes.FIELD_INT64);
      propTypeToProgress.put(ParmType.DEC, ProgressParserTokenTypes.FIELD_DEC);
      propTypeToProgress.put(ParmType.LOG, ProgressParserTokenTypes.FIELD_LOGICAL);
      propTypeToProgress.put(ParmType.DATE, ProgressParserTokenTypes.FIELD_DATE);
      propTypeToProgress.put(ParmType.DT, ProgressParserTokenTypes.FIELD_DATETIME);
      propTypeToProgress.put(ParmType.DTTZ, ProgressParserTokenTypes.FIELD_DATETIME_TZ);
      propTypeToProgress.put(ParmType.HANDLE, ProgressParserTokenTypes.FIELD_HANDLE);
      propTypeToProgress.put(ParmType.RECID, ProgressParserTokenTypes.FIELD_RECID);
      propTypeToProgress.put(ParmType.ROWID, ProgressParserTokenTypes.FIELD_ROWID);
   }

   /**
    * Get singleton instance of the helper.
    *
    * @return  singleton instance of the helper.
    */
   public static DynamicTablesHelper getInstance()
   {
      if (instance == null)
      {
         synchronized (DynamicTablesHelper.class)
         {
            if (instance == null)
            {
               instance = new DynamicTablesHelper();
               PatternEngine.setDebugLevel(ConversionDriver.MSG_NONE);
            }
         }
      }

      return instance;
   }

   /**
    * Generates DMO interface, implementation class and mapping document. TODO
    *
    * @param builder
    *        Temporary table builder containing definitions of the target table.
    *
    * @return <code>true</code> on success.
    */
   public boolean generateTable(TempTableBuilder builder)
   {
      long tableId = getUniqueTableId();

      String tableName = String.format("dtt%d", tableId);
      String ifaceName = String.format("DynamicRecord%d", tableId);
      String implClassName = String.format("DynamicRecord%dImpl", tableId);
      Map<String, String> clsFields4GLtoHibernate = new HashMap<String, String>();

      long fieldId = 1;
      for (P2JField field : builder.getFields())
      {
         String fieldNameHibernate = String.format("field%d", fieldId++);
         clsFields4GLtoHibernate.put(field.getName(), fieldNameHibernate);
      }

      Aast root = generateDMOAst(builder, ifaceName, tableName, clsFields4GLtoHibernate);

      try
      {
         PatternEngine engine;
         engine = new PatternEngine();
         engine.addTargetAst(root);
         engine.run("schema/p2o");

         root = (Aast) engine.getStoredObject(null);
         root.putAnnotation("permanent", false);
         root.putAnnotation("package", DMO_BASE_PACKAGE);

         // generate .hbm.xml
         engine = new PatternEngine();
         engine.addTargetAst(root);
         engine.run("schema/hibernate");
         XmlAst hbmXml = (XmlAst) engine.getStoredObject(ifaceName);

         // generate interface and implementation Java ASTs
         engine = new PatternEngine();
         engine.addTargetAst(root);
         engine.run("schema/java_dmo");
         Aast impl = (Aast) engine.getStoredObject(DMO_IMPL_CATEGORY, ifaceName);
         Aast iface = (Aast) engine.getStoredObject(DMO_IFACE_CATEGORY, ifaceName);

         // generate Java code from Java ASTs
         engine = new PatternEngine();
         engine.addTargetAst(impl);
         engine.addTargetAst(iface);
         engine.run("convert/brew");

         ByteArrayOutputStream implCodeStream =
                              (ByteArrayOutputStream) engine.getStoredObject(implClassName);
         ByteArrayOutputStream ifaceCodeStream =
                              (ByteArrayOutputStream) engine.getStoredObject(ifaceName);

         String fullIfaceName = String.format("%s.%s.%s",
                                              DMO_BASE_PACKAGE,
                                              DatabaseManager.TEMP_TABLE_SCHEMA,
                                              ifaceName);
         String fullImplClassName = String.format("%s.%s.impl.%s",
                                                  DMO_BASE_PACKAGE,
                                                  DatabaseManager.TEMP_TABLE_SCHEMA,
                                                  implClassName);
         StringBuilder sb = new StringBuilder();
         Map<String, Class<?>> classes = InMemoryCompiler.compile(new String[]
                                                                  {
                                                                     fullIfaceName,
                                                                     fullImplClassName
                                                                  },
                                                                  new String[]
                                                                  {
                                                                      ifaceCodeStream.toString(),
                                                                      implCodeStream.toString()
                                                                  },
                                                                  sb, true);

         if (classes == null)
         {
            //TODO
            ErrorManager.displayError(sb.toString());
            return false;
         }

         Class dmoIface = classes.get(fullIfaceName);
         Class dmoClass = classes.get(fullImplClassName);
         Class bufClass = dmoIface.getDeclaredClasses()[0];

         DMOIndex.registerDynamicTable(dmoIface, dmoClass);
         DatabaseManager.registerDynamicTable(dmoIface, dmoClass);

         Configuration cfg =
               DatabaseManager.getConfigBySchema(DatabaseManager.TEMP_TABLE_SCHEMA);
         cfg.addDocument(XmlAst.astToDom(hbmXml));
         cfg.buildMappings();
         PersistentClass pc = cfg.getClassMapping(dmoClass.getName());
         MutableSessionFactoryImpl sessionFactory =
               (MutableSessionFactoryImpl) DatabaseManager.getSessionFactory(DatabaseManager.TEMP_TABLE_DB);
         sessionFactory.add(pc, cfg);

         builder.setBufClass(bufClass);
         fields4GLtoHibernate.get().put(bufClass, clsFields4GLtoHibernate);
      }
      catch (Exception e)
      {
         //TODO
         ErrorManager.displayError(e.getMessage());
         return false;
      }

      return true;
   }

   /**
    * Generate and return a new unique id.
    *
    * @return  a new unique id.
    */
   private long getUniqueTableId()
   {
      return tableCounter.getAndIncrement();
   }

   /**
    * Generate AST that represents the target table and has the same structure as .p2o files from
    * the conversion stage.
    *
    * @param  builder
    *         Temporary table builder containing definitions of the target table.
    * @param  ifaceName
    *         The name of the target DMO interface.
    * @param  tableName
    *         The name of the target SQL table.
    * @param  fields4GLtoHibernate
    *         Map 4GL field names to Hibernate names.
    *
    * @return the root node of the generated AST.
    */
   private Aast generateDMOAst(TempTableBuilder builder,
                               String ifaceName,
                               String tableName,
                               Map<String, String> fields4GLtoHibernate)
   {
      Aast root = new ProgressAst(new CommonToken(ProgressParserTokenTypes.DATABASE,
                                                  DatabaseManager.TEMP_TABLE_SCHEMA));

      ProgressAst cls = new ProgressAst(new CommonToken(ProgressParserTokenTypes.TABLE,
                                                        builder.getTableName()));
      cls.putAnnotation("class-name", ifaceName);
      cls.putAnnotation("table-name", tableName);
      root.addChild(cls);

      for (P2JField field : builder.getFields())
      {
         String fieldName = fields4GLtoHibernate.get(field.getName());
         ParmType type = field.getType();
         int progressType = propTypeToProgress.get(type);
         ProgressAst property = new ProgressAst(new CommonToken(progressType, fieldName));
         property.putAnnotation("format", field.getFormat());
         property.putAnnotation("label", field.getLabel());

         long extent = field.getExtent();
         if (extent > 1)
            property.putAnnotation("extent", extent);

         cls.addChild(property);

         BaseDataType initial = field.getInitial();
         if (initial != null)
         {
            ProgressAst initialNode = new ProgressAst(
                                      new CommonToken(ProgressParserTokenTypes.KW_INIT, ""));
            property.addChild(initialNode);

            int valType;
            if (initial.isUnknown())
               valType = ProgressParserTokenTypes.UNKNOWN_VAL;
            else switch (type)
            {
               case CHAR:
                  valType = ProgressParserTokenTypes.STRING;
                  break;
               case INT:
                  valType = ProgressParserTokenTypes.NUM_LITERAL;
                  break;
               case DEC:
                  valType = ProgressParserTokenTypes.DEC_LITERAL;
                  break;
               case LOG:
                  valType = initial.equals(new logical(true)) ?
                            ProgressParserTokenTypes.BOOL_TRUE :
                            ProgressParserTokenTypes.BOOL_FALSE;
                  break;
               default:
                  throw new IllegalStateException();
            }

            String valStr;
            switch (type)
            {
               case CHAR:
                  valStr = new StringBuilder().append("\"").
                                               append(initial.toStringMessage()).
                                               append("\"").toString();
                  break;
               default:
                  valStr = initial.toStringMessage();
            }

            ProgressAst initialVal = new ProgressAst(
                                        new CommonToken(valType, valStr));
            initialNode.addChild(initialVal);
         }
      }

      assignIds(root, 1);

      return root;
   }

   /**
    * Traverse an AST and assign some unique ids to all of its nodes.
    *
    * @param root
    *        The root node of the AST.
    */
   private long assignIds(Aast root, long startId)
   {
      root.setId(startId++);

      int childCount = root.getNumImmediateChildren();
      if (childCount > 0)
      {
         for (int i = 0; i < childCount; i++)
         {
            Aast child = root.getChildAt(i);
            startId = assignIds(child, startId);
         }
      }
      return startId;
   }
}
